/* This is logoEPIwm */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>

#ifndef NOXPM
#include <X11/xpm.h>
#endif

#include <stdio.h>
#include <string.h>

#include "logoEPIwm.xbm"
#include "logoEPIwm_mask.xbm"
#include "logoEPIwm_icon.xbm"

#ifndef NOXPM
#include "logoEPIwm.xpm"
#endif

#include "patchlevel.h"


static char	*progname;
Display		*display;
int		screen_num;

void main(argc, argv)
     int argc;
     char **argv;
{
  /* Display, window and gc manipulation variables */
  Window		win;
  GC			gc;
  XSetWindowAttributes	setwinattr;
  XGCValues		gcvalues;
  unsigned long		valuemask, gcvaluemask, inputmask;
  unsigned int		width, height;
  int			x, y, geomflags;
  unsigned int		border_width = 0;
  unsigned int		display_width, display_height, display_depth;
  char			*window_name = "logoEPIwm";
  char			*icon_name = "logoEPIwm";
  Pixmap		icon_pixmap, background_pixmap, shape_pixmap;
  XSizeHints		size_hints;
  XIconSize		*size_list;
  XWMHints		wm_hints;
  XClassHint		class_hints;
  XTextProperty		windowName, iconName;
  int			count, argnum;
  int			use_wm, float_up, use_mono;
  XEvent		report;
  char			*display_name = NULL;
  char			buffer[20];
  int			bufsize = 20;
  KeySym		keysym;
  XComposeStatus	compose;
  int			charcount;
  Cursor		cursor;
  time_t		time1;
  time_t		time2;
  int			eternal;
  
#ifndef NOXPM
  /* Color allocation variables */
  Visual	*default_visual;
  Colormap	default_cmap;
  XpmAttributes	xpmattributes;
  XVisualInfo	*visual_info, vinfo_template;
  int		nmatches;
  static char	*visual_name[]={ "StaticGray", "GrayScale", "StaticColor",
				  "PseudoColor", "TrueColor", "DirectColor" };
#endif

  /* Window movement variables */
  XWindowChanges	winchanges;
  Window		root, child, basewin;
  int			offs_x, offs_y, new_x, new_y, tmp_x, tmp_y;
  unsigned int		tmp_mask;
  
  progname = argv[0];

  /* Option handling: "-w", "-f" "-mono" and "-geometry" are recognized */
  use_wm = False;
  float_up = False;
  use_mono = False;
  x = y = 0;
  geomflags = 0;
  eternal = 0;
  for(argnum=1; argnum<argc; argnum++)
    {
      if (!strcmp(argv[argnum],"-w"))
	use_wm = True;
      if (!strcmp(argv[argnum],"-f"))
	float_up = True;
      if (!strcmp(argv[argnum],"-mono"))
	use_mono = True;
      if (!strcmp(argv[argnum],"-e"))
	eternal = 1;
      if (!strcmp(argv[argnum],"-geometry"))
	geomflags = XParseGeometry(argv[++argnum], &x, &y, &width, &height);
    }
  
  /* Connect to X server */
  if ( (display = XOpenDisplay(display_name)) == NULL )
    {
      (void) fprintf(stderr, "%s: Cannot connect to X server %s\n",
		     progname, XDisplayName(display_name));
      exit(-1);
    }

  /* Get screen size and depth */
  screen_num = DefaultScreen(display);
  display_width = DisplayWidth(display, screen_num);
  display_height = DisplayHeight(display, screen_num);
  display_depth = DefaultDepth(display, screen_num);

  /* Set the window size to snugly fit the teddybear pixmap */
  width = logoEPIwm_width;
  height = logoEPIwm_height;

  /* Set the window position according to user preferences */
  if (geomflags & XNegative)
    x = display_width - width + x;
  if (geomflags & YNegative)
    y = display_height - height + y;
  /* Clip against bounds to stay on the screen */
  if (x<0) x=0;
  if (x > display_width - width) x = display_width - width;
  if (y<0) y=0;
  if (y > display_height - height) y = display_height - height;

  /* Create the main window */
  win = XCreateSimpleWindow(display, RootWindow(display,screen_num),
			    (display_width-width)/2, 
			    (display_height-height)/2,
			    width,height,border_width,
			    BlackPixel(display,screen_num),
			    WhitePixel(display,screen_num));
  basewin = win;

  /* Create a GC (Currently not used for any drawing) */
  gcvalues.foreground = BlackPixel(display,screen_num);
  gcvalues.background = WhitePixel(display,screen_num);
  gcvaluemask = GCForeground | GCBackground;
  gc = XCreateGC(display, win, gcvaluemask, &gcvalues);

#ifdef NOXPM
  /* Use b/w dithered X bitmap no matter what */
  background_pixmap =
    XCreatePixmapFromBitmapData(display, win, logoEPIwm_bits,
				logoEPIwm_width, logoEPIwm_height,
				BlackPixel(display, screen_num),
				WhitePixel(display, screen_num),
				display_depth);
#else
  /* Get information about the default visual */
  default_visual = DefaultVisual(display, screen_num);
  default_cmap = DefaultColormap(display, screen_num);

  /* Get the visual class of the default visual. (I strongly feel */
  /* that there must be a more straightforward way to do this...) */
  vinfo_template.visualid = XVisualIDFromVisual(default_visual);
  visual_info = XGetVisualInfo(display, VisualIDMask,
			       &vinfo_template, &nmatches);
  /* The default visual should be supported, so don't check for NULL return */
#ifdef DEBUG
  printf("%s: default visual class is %s.\n",
	 progname, visual_name[visual_info->class]);
#endif
  if ((visual_info->class == StaticGray) || (use_mono))
    {
      /* Use b/w dithered bitmap */
      background_pixmap =
	XCreatePixmapFromBitmapData(display, win, logoEPIwm_bits,
				    logoEPIwm_width, logoEPIwm_height,
				    BlackPixel(display, screen_num),
				    WhitePixel(display, screen_num),
				    display_depth);
    }
  else /* at least GrayScale - let Xpm decide which visual to use */
    {
      xpmattributes.visual = default_visual;
      xpmattributes.colormap = default_cmap;
      xpmattributes.depth = display_depth;
      xpmattributes.valuemask = XpmVisual | XpmColormap | XpmDepth;
      if (XpmCreatePixmapFromData(display, win, logoEPIwm,
				  &background_pixmap, &shape_pixmap,
				  &xpmattributes) < XpmSuccess)
	{
	  printf("%s: Failed to allocate colormap. Using black-and-white.\n",
		 progname);
	  background_pixmap =
	    XCreatePixmapFromBitmapData(display, win, logoEPIwm_bits,
					logoEPIwm_width, logoEPIwm_height,
					BlackPixel(display, screen_num),
					WhitePixel(display, screen_num),
					display_depth);
	}
    }
  XFree(visual_info);
#endif
  setwinattr.background_pixmap = background_pixmap;
  if (use_wm)
    setwinattr.override_redirect = False;
  else
    setwinattr.override_redirect = True;
  cursor = XCreateFontCursor(display, XC_fleur);
  setwinattr.cursor = cursor;
  valuemask = CWBackPixmap | CWOverrideRedirect | CWCursor;
  XChangeWindowAttributes(display, win, valuemask, &setwinattr);

  /* Create and set the shape pixmap of the window - requires shape Xext */
  shape_pixmap = XCreateBitmapFromData(display,win, logoEPIwm_mask_bits,
				      logoEPIwm_mask_width, logoEPIwm_mask_height);
  XShapeCombineMask(display, win, ShapeBounding, 0, 0, shape_pixmap, ShapeSet);

  /* Get available icon sizes from window manager */
  /* (and then blatantly ignore the result)       */
  if (XGetIconSizes(display, RootWindow(display,screen_num),
		    &size_list, &count) == 0)
    {
      /* Window manager didn't set preferred icon sizes - use the default */
      icon_pixmap = XCreateBitmapFromData(display, win, logoEPIwm_icon_bits, logoEPIwm_icon_width, logoEPIwm_icon_height);
      /*      icon_pixmap = XCreatePixmapFromBitmapData(display, win, logoEPIwm_icon_bits, logoEPIwm_icon_width, logoEPIwm_icon_height, BlackPixel(display, screen_num), WhitePixel(display, screen_num), display_depth);*/
    }
  else
    {
      /* Ignore the list and use the default size anyway */
      icon_pixmap = XCreateBitmapFromData(display, win, logoEPIwm_icon_bits, logoEPIwm_icon_width, logoEPIwm_icon_height);
      /*      icon_pixmap = XCreatePixmapFromBitmapData(display, win, logoEPIwm_icon_bits, logoEPIwm_icon_width, logoEPIwm_icon_height, BlackPixel(display, screen_num), WhitePixel(display, screen_num), display_depth); */
    }
  /* Report size hints and other stuff to the window manager */
  size_hints.min_width = width;    /* Don't allow any resizing */
  size_hints.min_height = height;
  size_hints.max_width = width;
  size_hints.max_height = height;
  size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
  if (XStringListToTextProperty(&window_name, 1, &windowName) == 0)
    {
      (void) fprintf(stderr,
		     "%s: structure allocation for windowName failed.\n",
		     progname);
      exit(-1);
    }
  if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0)
    {
      (void) fprintf(stderr,
		     "%s: structure allocation for iconName failed.\n",
		     progname);
      exit(-1);
    }
  wm_hints.initial_state = NormalState;
  wm_hints.input = True;
  wm_hints.icon_pixmap = icon_pixmap;
  wm_hints.flags = StateHint | IconPixmapHint | InputHint;
  
  class_hints.res_name = progname;
  class_hints.res_class = "logoEPIwm";
  
  XSetWMProperties(display, win, &windowName, &iconName,
		   argv, argc, &size_hints, &wm_hints, &class_hints);

  /* Select event types wanted */
  inputmask = ExposureMask | KeyPressMask | ButtonPressMask | 
    ButtonReleaseMask | StructureNotifyMask | ButtonMotionMask | 
      PointerMotionHintMask | EnterWindowMask | LeaveWindowMask;
  if (float_up) inputmask |= VisibilityChangeMask;
  XSelectInput(display, win, inputmask);

  /* Display window */
  XMapWindow(display,win);

  /* Get and process the events */
  while (1)
    {
      XNextEvent(display, &report);
      switch(report.type)
	{
	case Expose:
	  if (report.xexpose.count != 0)
	    break;
	  else
	    {
	      /* No drawing needed - the background pixmap */
	      /* is handled automatically by the X server  */
	    }
	  break;
	case ConfigureNotify:
	  /* Window has been resized */
	  width = report.xconfigure.width;
	  height = report.xconfigure.height;
	  break;
	case ReparentNotify:
	  /* Window was reparented by the window manager */
	  if (!use_wm)
	    (void) fprintf(stderr,
			   "%s: Window manager wouldn't leave window alone!\n",
			   progname);
	  basewin = report.xreparent.parent;
	  break;
	case EnterNotify:
	  /* Grab the keyboard while the pointer is in the window */
	  XGrabKeyboard(display, win, False, GrabModeAsync, GrabModeAsync,
			CurrentTime);
	  break;
	case LeaveNotify:
	  /* Release the keyboard when the pointer leaves the window */
	  XUngrabKeyboard(display, CurrentTime);
	  break;
	case ButtonPress:
	  /* Raise logoEPIwm above sibling windows  */
	  XRaiseWindow(display, win);
	  /* Remember where the mouse went down */
	  XQueryPointer(display, basewin, &root, &child, &tmp_x, &tmp_y,
			&offs_x, &offs_y, &tmp_mask);
	  break;
	case ButtonRelease:
	  /* Place logoEPIwm at the new position */
	  XQueryPointer(display, basewin, &root, &child, &new_x, &new_y,
		        &tmp_x, &tmp_y, &tmp_mask);
	  winchanges.x = new_x - offs_x;
	  winchanges.y = new_y - offs_y;
	  XReconfigureWMWindow(display, basewin, screen_num,
			       CWX | CWY, &winchanges);
	  break;
	case MotionNotify:
	  /* Move logoEPIwm around with the mouse */
	  while (XCheckMaskEvent(display, ButtonMotionMask, &report));
	  if (!XQueryPointer(display, report.xmotion.window, &root, &child,
			    &new_x, &new_y, &tmp_x, &tmp_y, &tmp_mask))
	    break;
	  winchanges.x = new_x - offs_x;
	  winchanges.y = new_y - offs_y;
	  XReconfigureWMWindow(display, win, screen_num,
			       CWX | CWY, &winchanges);
	  break;
	case VisibilityNotify:
	  /* Put logoEPIwm on top of overlapping windows */
	  if (float_up)
	    if ((report.xvisibility.state == VisibilityFullyObscured)
		|| (report.xvisibility.state == VisibilityPartiallyObscured))
	      XRaiseWindow(display,win);
	  break;
	case KeyPress:
	  /* Exit on "q" or "Q" */
	  /*
	  **charcount = XLookupString(&report.xkey, buffer, bufsize,
	  **		    &keysym, &compose);
	  **if((keysym == XK_Q) || (keysym == XK_q))
	  **{
	  **  XCloseDisplay(display);
	  **  exit(1);
	  **  }
	  */
	  XCloseDisplay(display);
	  exit(1);
	  break;
	default:
	  /* Throw away all other events */
	  break;
	} /* end switch */
      if (!eternal)
	{
	  time1 = time(NULL);
	  time2 = time(NULL);
	  while (difftime(time2, time1) < 2)
	    time2 = time(NULL);
	  exit(1);
	}
    } /* end while */
}
